use std::{
    ffi::CString,
    os::{
        fd::OwnedFd,
        unix::io::{AsRawFd, RawFd},
    },
    ptr,
};

use libseccomp::ScmpFilterContext;
use nix::{
    self,
    errno::Errno,
    libc::c_char,
    sys::wait::{waitpid, WaitStatus},
    unistd::{close, fork, setpgid, ForkResult, Pid},
};

use crate::unshare::{child, config::Config, error::Error, Child, Command, ExitStatus};

type ChildPreExecFunc = Box<dyn Fn() -> Result<(), Errno>>;

pub struct ChildInfo<'a> {
    pub filename: *const c_char,
    pub args: &'a [*const c_char],
    pub cfg: &'a Config,
    pub pre_exec: &'a Option<ChildPreExecFunc>,
    pub seccomp_filter: &'a Option<ScmpFilterContext>,
    pub seccomp_sender: Option<&'a OwnedFd>,
}

fn raw_with_null(arr: &Vec<CString>) -> Vec<*const c_char> {
    let mut vec = Vec::with_capacity(arr.len().saturating_add(1));
    for i in arr {
        vec.push(i.as_ptr());
    }
    vec.push(ptr::null());
    vec
}

impl Command {
    /// Run the command and return exit status
    pub fn status(&mut self) -> Result<ExitStatus, Error> {
        // TODO(tailhook) stdin/stdout/stderr
        self.spawn()?
            .wait()
            .map_err(|e| Error::WaitError(e.raw_os_error().unwrap_or(-1)))
    }

    /// Spawn the command and return a handle that can be waited for
    pub fn spawn(&mut self) -> Result<Child, Errno> {
        let c_args = raw_with_null(&self.args);

        // Clear the environment.
        std::env::remove_var("SYD_INIT");
        std::env::remove_var(crate::config::ENV_LOG);
        std::env::remove_var(crate::config::ENV_NO_SYSLOG);
        std::env::remove_var(crate::config::ENV_NO_CROSS_MEMORY_ATTACH);
        std::env::remove_var(crate::config::ENV_FAKEROOT);
        std::env::remove_var(crate::config::ENV_SH);
        std::env::remove_var(crate::config::ENV_UNSHARE_MOUNT);
        std::env::remove_var(crate::config::ENV_UNSHARE_UTS);
        std::env::remove_var(crate::config::ENV_UNSHARE_IPC);
        std::env::remove_var(crate::config::ENV_UNSHARE_USER);
        std::env::remove_var(crate::config::ENV_UNSHARE_PID);
        std::env::remove_var(crate::config::ENV_UNSHARE_NET);
        std::env::remove_var(crate::config::ENV_UNSHARE_CGROUP);

        // SAFETY: In libc we trust.
        match unsafe { fork() }? {
            ForkResult::Parent { child, .. } => {
                let seccomp_fd = match self.after_start(child) {
                    Ok(seccomp_fd) => seccomp_fd,
                    Err(e) => loop {
                        match waitpid(child, None) {
                            Ok(WaitStatus::Exited(_, errno)) => return Err(Errno::from_i32(errno)),
                            Err(Errno::EINTR) => {}
                            _ => return Err(e),
                        }
                    },
                };

                Ok(Child {
                    pid: child.into(),
                    status: None,
                    seccomp_fd,
                })
            }
            ForkResult::Child => {
                let seccomp_sender = self.seccomp_channel.as_ref().map(|c| c.channel());
                let seccomp_sender = if let Some((sender, receiver)) = &seccomp_sender {
                    // The receiver is unneeded, close the underlying fd.
                    let _ = close(receiver.fd.as_raw_fd());
                    Some(sender.fd)
                } else {
                    None
                };
                let child_info = ChildInfo {
                    filename: self.filename.as_ptr(),
                    args: &c_args[..],
                    cfg: &self.config,
                    pre_exec: &self.pre_exec,
                    seccomp_filter: &self.seccomp_filter,
                    seccomp_sender,
                };
                child::child_after_clone(&child_info);
            }
        }
    }

    #[allow(clippy::cognitive_complexity)]
    fn after_start(&mut self, pid: Pid) -> Result<Option<RawFd>, Errno> {
        // Receive the seccomp notification file descriptor.
        let seccomp_fd = if let Some(ref c) = self.seccomp_channel {
            let (sender, receiver) = c.channel();
            let _ = close(sender.fd.as_raw_fd());
            match receiver.recvfd() {
                Ok(fd) => {
                    let _ = close(receiver.fd.as_raw_fd());
                    Some(fd)
                }
                Err(Errno::EINVAL) => {
                    // seccomp filter failed to load.
                    return Err(Errno::EBUSY);
                }
                Err(errno) => {
                    return Err(errno);
                }
            }
        } else {
            None
        };

        if self.config.make_group_leader {
            setpgid(pid, pid)?;
        }

        if let Some(ref mut callback) = self.before_unfreeze {
            #[allow(clippy::cast_sign_loss)]
            callback(i32::from(pid) as u32)?;
        }

        Ok(seccomp_fd)
    }
}
